pro do_map_subarrays, map_subarr, weight_subarr, file_scalars=file_scalars, ind_scans=ind_scans, $
                  ind_subscans=ind_subscans, maxnoise=maxnoise, comp_errors=comp_errors, error_subarr


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


restore, file_scalars

restore, file_pixcoords

restore, file_signal


dimxy = dimx * dimy

nb_row = 16L
nb_col = 18L
nb_mat = nb_row * nb_col
nmat = nb / nb_mat
if nmat eq 8L then nsub_x = 4L else nsub_x = 2L
nb_line = nsub_x * nb_row

map_subarr = fltarr(dimx, dimy, nmat)
weight_subarr = fltarr(dimx, dimy, nmat)
if keyword_set(comp_errors) then error_subarr = fltarr(dimx, dimy, nmat)



weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))

tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]


indimprint = nstep_imprint * floor(nstep_imprint * (y_bolos - floor(y_bolos))) $
   + floor(nstep_imprint * (x_bolos - floor(x_bolos)))

;;;restore, file_imprint
;;;
pixsize = pix_as
diameter = 2. * (1.2 * pix_as) / sqrt(!pi)
nstep_imprint = max([nstep_imprint, 2 * ceil(2.5 * pixsize / diameter)])
   model_r_as = findgen(10000L)/ 100.
   model_imprint = fltarr(10000L)
   w = where(model_r_as le diameter / 2.)
   model_imprint(w) = 1.
   pix_cen = dim_imprint / 2
   np = nstep_imprint^2
   imprint = fltarr(np, dim_imprint, dim_imprint)
   xc_imprint = fltarr(np)
   yc_imprint = fltarr(np)
   for i = 0L, nstep_imprint - 1 do begin
      xc_imprint(i*nstep_imprint:(i+1)*nstep_imprint-1) = findgen(nstep_imprint)
      yc_imprint(i*nstep_imprint:(i+1)*nstep_imprint-1) = i
   endfor
   xc_imprint = xc_imprint / float(nstep_imprint) + pix_cen
   yc_imprint = yc_imprint / float(nstep_imprint) + pix_cen
   dim_fine = dim_imprint * nstep_imprint
   tab_x_fine = 0.5 + (fltarr(dim_fine) + 1.) ## findgen(dim_fine)
   tab_y_fine = 0.5 + findgen(dim_fine) ## (fltarr(dim_fine) + 1.)
   xc_fine = xc_imprint * nstep_imprint
   yc_fine = yc_imprint * nstep_imprint
   scale = pixsize / float(nstep_imprint)
   for i = 0L, np - 1 do begin
      ray_as = sqrt((tab_x_fine - xc_fine(i))^2. + (tab_y_fine - yc_fine(i))^2.) * scale
      beam_fine = interpol(model_imprint, model_r_as, ray_as)
      beam = rebin(beam_fine, dim_imprint, dim_imprint)
      imprint(i, *, *) = beam / total(beam)
   endfor
beam = 0
beam_fine = 0
tab_x_fine = 0
tab_y_fine = 0
xc_imprint = 0
yc_imprint = 0
model_imprint = 0
model_r_as = 0
;;;

;;;dim_imprint = long(n_elements(imprint(0, 0, *)))
sq_dim_imprint = dim_imprint^2
imprint_reform = reform(imprint, nstep_imprint^2, sq_dim_imprint)
imprint = 0


if !version.memory_bits ge 64 then indxy_beam = -9L + lon64arr(nt, nb) else indxy_beam = -9L + lonarr(nt, nb)

for mat = 0, nmat - 1 do begin
   wmat = reform(rebin((mat mod nsub_x) * nb_row + lindgen(nb_row), nb_row, nb_col) $
         + nb_line * (((mat / nsub_x) * nb_col + lindgen(nb_col)) ## (1L + lonarr(nb_row))), nb_mat)
   indxy_beam(*, *) = -9L
   slice_indxy = floor(x_bolos(*, wmat)) + dimx * floor(y_bolos(*, wmat))

   ;;; suppress invalid samples:
   wnan = where(finite(signal(*, wmat)) lt 1 $
      or x_bolos(*, wmat) lt xcmin_imprint or x_bolos(*, wmat) ge dimx - xcmin_imprint $
      or y_bolos(*, wmat) lt xcmin_imprint or y_bolos(*, wmat) ge dimy - xcmin_imprint, cw_nan)
   if cw_nan ge 1 then slice_indxy(wnan) = -9L
   wnan = 0

   indxy_beam(*, wmat) = slice_indxy

   histo_indxy = histogram(indxy_beam, reverse_indices=revind_fine, min=0., max=dimxy-1)
   wproj_fine = where(histo_indxy ge 3L, nproj_fine)
   histo_indxy = 0

   ;;; 1D index array of imprint lower-left corner (repetitions eliminated by wproj_fine):
   vectpos_startimprint = floor(-xcmin_imprint + (wproj_fine mod dimx)) $
      + dimx * floor(-xcmin_imprint + wproj_fine / dimx)

   ;;; 1D index array of entire imprint, relative to the lower-left corner:
   tab_ind = lindgen(dim_imprint)
   sq_dim_imprint = dim_imprint^2
   vectpos_imprint = reform(rebin(tab_ind, dim_imprint, dim_imprint), sq_dim_imprint) $
      + dimx * rebin(tab_ind, sq_dim_imprint, /sample)



   map_weight = fltarr(dimx, dimy)
   map_signal = fltarr(dimx, dimy)
   if keyword_set(comp_errors) then begin
      map_error = fltarr(dimx, dimy)
      map_sqw = fltarr(dimx, dimy)
   endif

   for k = 0L, nproj_fine - 1L do begin
      revind_0 = revind_fine(wproj_fine(k))
      revind_1 = revind_fine(wproj_fine(k)+1)
      cp = revind_1 - revind_0
      wp = revind_fine(revind_0:revind_1-1)
      wt = wp mod nt

      vectpos = vectpos_startimprint(k) + vectpos_imprint
      bin_weight = rebin(weight(tab_subscan(wt) + nsubscans * (wp / nt)), cp, sq_dim_imprint) $
         * imprint_reform(indimprint(wp), *)
      bin_signal = rebin(signal(wp), cp, sq_dim_imprint)
      map_weight(vectpos) += total(bin_weight, 1)
      map_signal(vectpos) += total(bin_signal * bin_weight, 1)
      if keyword_set(comp_errors) then begin
         map_sqw(vectpos) += total(bin_weight^2., 1)
         map_error(vectpos) += total(bin_signal^2. * bin_weight, 1)
      endif
   endfor

   wcov = where(map_weight gt 0.)
   if keyword_set(comp_errors) then begin
      map_error(wcov) = (map_error(wcov) * map_weight(wcov) - map_signal(wcov)^2.) / (map_weight(wcov)^2. - map_sqw(wcov))
      map_error(wcov) = sqrt(map_error(wcov) / map_weight(wcov))
   endif
   map_signal(wcov) /= map_weight(wcov)


   map_subarr(*, *, mat) = map_signal
   weight_subarr(*, *, mat) = map_weight
   if keyword_set(comp_errors) then error_subarr(*, *, mat) = map_error
endfor


end
