pro ingest_data_artemis, file_param_blocks, ind_scans, ind_subscans, nsubscans, vscan_as_s, timestep, $
                 narr=narr, nscans=nscans, nt=nt, nblocks=nblocks, block=block, $
                 inst=inst, nbolo=nbolo, boloref=boloref, ind_arrays_eff=ind_arrays_eff, $
                 arrays_default=arrays_default, xdeg_blocks=xdeg_blocks, ydeg_blocks=ydeg_blocks, $
                 orient=orient, timegap=timegap, dir_in=dir_in, files_scans=files_scans, $
                 hdr_ref=hdr_ref, cutout=cutout, file_XYscan=file_XYscan, file_times=file_times, file_vscan=file_vscan, $
                 root_signal=root_signal, root_xyas=root_xyas, maxtimediff=maxtimediff, $
                 visu=visu, debug=debug, vis_traject=vis_traject, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, $
                 vs_min = vs_min, vs_max=vs_max, spiral=spiral, nzadata=nzadata, elec_boxes=elec_boxes


;;; creates signal, coordinates, scan speed and non-zero acceleration flag arrays
;;; computes time-related quantities

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



crs = string(13b) + string(10b)

cut_margin = 0.05
nt_scan_min = 5000L

restore, file_param_blocks
restore, dir_in + files_scans(0)
scan.flag_scanleg = abs(scan.flag_scanleg)

;;;
nscans_block = nscans
i_start_cross = 1

cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)
if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   p_cos_scanrot = cos(p_scanrot_deg * !dpi / 180.D0)
   p_sin_scanrot = sin(p_scanrot_deg * !dpi / 180.D0)
endif
;;;


if nblocks gt 1 then begin
   print, crs + 'building block ' + chain(block+1) + ' / ' + chain(nblocks)
   print, crs
endif

w = where(ind_arrays_eff eq 1)
ind = w(0)
elec_boxes = !values.f_nan + fltarr(nscans, nbolo(ind))


finish_init = 1
repeat begin
   nt = n_elements(scan.time_s)
   tags = tag_names(scan)
   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(*, boloref(ind))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(*, boloref(ind))

   if keyword_set(p_mid_Xra) then if (abs(p_mid_Xra - mid_Xra) gt cos(mid_Ydec * !dpi / 180.D0) / 60.D0 $
    or abs(p_mid_Ydec - mid_Ydec) gt 1.D0 / 60.D0) then begin
      p_Xra = Xra
      p_Ydec = Ydec
      gnomonic, p_Xra, p_Ydec, refx=p_mid_Xra, refy=p_mid_Ydec
   endif

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
   if keyword_set(offset_ra_as) then begin
      Xra -= (offset_ra_as(0) / 3600.D0)
      Ydec += (offset_dec_as(0) / 3600.D0)
   endif
   if keyword_set(cutout) then begin
      wcut = where(sqrt(Xra^2. + Ydec^2.) le cutout(2) + cut_margin, nt)
      Xra = Xra(wcut)
      Ydec = Ydec(wcut)
   endif

   if nblocks gt 1 then begin
      Xra_scan = Xra * cos_scanrot + Ydec * sin_scanrot + off_cen_Xra
      Ydec_scan = Ydec * cos_scanrot - Xra * sin_scanrot + off_cen_Ydec
      wtblock = where(finite(Xra) gt 0 and Xra_scan ge xdeg_blocks(block, 0) and Xra_scan le xdeg_blocks(block, 1) $
         and Ydec_scan ge ydeg_blocks(block, 0) and Ydec_scan le ydeg_blocks(block, 1), nt)
       if nt lt max([nt_scan_min, n_elements(scan.time_s) / 4L]) then begin
          nscans_block -= 1
          files_scans(i_start_cross-1) = ''
          restore, dir_in + files_scans(i_start_cross)
          scan.flag_scanleg = abs(scan.flag_scanleg)
          i_start_cross += 1
          finish_init = 0
      endif else finish_init = 1
   endif
endrep until finish_init eq 1


;;; attempt to eliminate initial slew:
if (not keyword_set(spiral)) then begin
   ntoff = min([100, nt / 20])
   dx = Xra(ntoff:nt-1) - Xra(0:nt-1-ntoff)
   dy = Ydec(ntoff:nt-1) - Ydec(0:nt-1-ntoff)
   vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
   dx = 0
   dy = 0
   med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
   if nt gt 12 * 100 then begin
      if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endif else begin
      if median(vect_scanrot_deg(0:nt/2)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endelse
   histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
   m = max(histo, ind)
   w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
   scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)

   w_exc_cross = 0
   if nblocks eq 1 then begin
      vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
      w = where(vect_diff le 10. or vect_diff ge 170.)
      if w(0) gt 0 then w_exc_cross = lindgen(w(0))
      w = 0
   endif
endif else w_exc_cross = 0

if keyword_set(w_exc_cross) then Xra(w_exc_cross) = !values.f_nan
wtblock = where(finite(Xra) gt 0, nt)
if keyword_set(cutout) then wtblock = wcut(wtblock)

;;;
if keyword_set(w_exc_cross) then begin
   scan.flag_scanleg(w_exc_cross) = -abs(scan.flag_scanleg(w_exc_cross))
   save, file=dir_in+files_scans(i_start_cross-1), scan, /xdr
endif
;;;
if nblocks gt 1 then wtblock = where(finite(Xra) gt 0 and $
   Xra_scan ge xdeg_blocks(block, 0) and Xra_scan le xdeg_blocks(block, 1) $
   and Ydec_scan ge ydeg_blocks(block, 0) and Ydec_scan le ydeg_blocks(block, 1), nt)


;;; time vector in seconds:
time_s = scan.time_s(wtblock) - scan.time_s(wtblock(0))
tab_subscans = scan.flag_scanleg(wtblock)
;;;
if (not keyword_set(spiral)) then begin
   tab_scanrot_deg = [vect_scanrot_deg(0)+dblarr(ntoff/2), vect_scanrot_deg, $
      vect_scanrot_deg(nt-1-ntoff)+dblarr(ntoff-ntoff/2)]
   tab_scanrot_deg = tab_scanrot_deg(wtblock)
endif
;;;

if !version.memory_bits ge 64 then nt = long64(nt)



;;; create pointers to signals for all arrays (as dblarr(nt, nb)):

signals = ptrarr(narr, /allocate_heap)

tags = tag_names(scan)
for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   *signals(detarr) = float(scan.(w(0))(wtblock, 0L:nbolo(detarr)-1L))
   elec_boxes(i_start_cross-1, 0L:nbolo(detarr)-1L) = scan.mat_boxes - min(scan.mat_boxes, /nan)
endif



;;; map coordinates in arcsec:

;cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
;sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)

xas_bolos = ptrarr(narr, /allocate_heap)
yas_bolos = ptrarr(narr, /allocate_heap)

if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
;   p_cos_scanrot = cos(p_scanrot_deg * !dpi / 180.D0)
;   p_sin_scanrot = sin(p_scanrot_deg * !dpi / 180.D0)
   p_xas_bolos = ptrarr(narr, /allocate_heap)
   p_yas_bolos = ptrarr(narr, /allocate_heap)
endif


for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(wtblock, 0:nbolo(detarr)-1)
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(wtblock, 0:nbolo(detarr)-1)

   if keyword_set(p_Xra) then begin
      p_Xra = Xra
      p_Ydec = Ydec
      gnomonic, p_Xra, p_Ydec, refx=p_mid_Xra, refy=p_mid_Ydec
      p_Xra *= 3600.D0
      p_Ydec *= 3600.D0
   endif

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
   Xra *= 3600.D0
   Ydec *= 3600.D0
   if keyword_set(offset_ra_as) then begin
      Xra -= offset_ra_as(0)
      Ydec += offset_dec_as(0)
   endif

   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      if keyword_set(p_Xra) then begin
         *p_xas_bolos(detarr) = p_Xra * p_cos_scanrot + p_Ydec * p_sin_scanrot
         *p_yas_bolos(detarr) = p_Ydec * p_cos_scanrot - p_Xra * p_sin_scanrot
      endif else begin
         *p_xas_bolos(detarr) = Xra * p_cos_scanrot + Ydec * p_sin_scanrot
         *p_yas_bolos(detarr) = Ydec * p_cos_scanrot - Xra * p_sin_scanrot
      endelse
   endif

   case orient of
      'astro': begin
         *xas_bolos(detarr) = Xra
         *yas_bolos(detarr) = Ydec
         end
      'scan': begin
         *xas_bolos(detarr) = Xra * cos_scanrot + Ydec * sin_scanrot
         *yas_bolos(detarr) = Ydec * cos_scanrot - Xra * sin_scanrot
         end
   endcase
endif



;;; read cross-scan data:

ind_scans = [0L, nt]
;;;for i = 1, nscans - 1 do begin
for i = i_start_cross, nscans - 1 do begin
;;;
   restore, dir_in + files_scans(i)
   scan.flag_scanleg = abs(scan.flag_scanleg)
   nt_b = n_elements(scan.time_s)
   tags = tag_names(scan)

   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(*, boloref(ind))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(*, boloref(ind))

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
   if keyword_set(offset_ra_as) then begin
      Xra -= (offset_ra_as(i) / 3600.D0)
      Ydec += (offset_dec_as(i) / 3600.D0)
   endif
   if keyword_set(cutout) then begin
      wcut = where(sqrt(Xra^2. + Ydec^2.) le cutout(2) + cut_margin, nt_b)
      Xra = Xra(wcut)
      Ydec = Ydec(wcut)
   endif


   ;;; attempt to eliminate initial slew:
   if (not keyword_set(spiral)) then begin
      ntoff = min([100, nt_b / 20])
      dx = Xra(ntoff:nt_b-1) - Xra(0:nt_b-1-ntoff)
      dy = Ydec(ntoff:nt_b-1) - Ydec(0:nt_b-1-ntoff)
      vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
      dx = 0
      dy = 0
      med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
      if nt_b gt 12 * 100 then begin
         if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endif else begin
         if median(vect_scanrot_deg(0:nt_b/2)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endelse
      histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
      m = max(histo, ind)
      w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
      scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)

      w_exc_cross = 0
      if nblocks eq 1 then begin
         vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
         w = where(vect_diff le 10. or vect_diff ge 170.)
         if w(0) gt 0 then w_exc_cross = lindgen(w(0))
         w = 0
      endif

      if keyword_set(w_exc_cross) then Xra(w_exc_cross) = !values.f_nan
;;;
      if keyword_set(w_exc_cross) then begin
         scan.flag_scanleg(w_exc_cross) = -abs(scan.flag_scanleg(w_exc_cross))
         save, file=dir_in+files_scans(i), scan, /xdr
      endif
;;;
   endif

   if nblocks gt 1 then begin
      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         Xra_rot = Xra * p_cos_scanrot + Ydec * p_sin_scanrot + p_off_cen_Xra
         Ydec_rot = Ydec * p_cos_scanrot - Xra * p_sin_scanrot + p_off_cen_Ydec
      endif else begin
         Xra_rot = Xra * cos_scanrot + Ydec * sin_scanrot + off_cen_Xra
         Ydec_rot = Ydec * cos_scanrot - Xra * sin_scanrot + off_cen_Ydec
      endelse
      wtblock = where(finite(Xra) gt 0 and Xra_rot ge xdeg_blocks(block, 0) and Xra_rot le xdeg_blocks(block, 1) $
         and Ydec_rot ge ydeg_blocks(block, 0) and Ydec_rot le ydeg_blocks(block, 1), nt_b)
   endif else wtblock = where(finite(Xra) gt 0, nt_b)

;;;
   if (nblocks eq 1 or nt_b ge max([nt_scan_min, n_elements(scan.time_s) / 4L])) then begin
;;;
   if keyword_set(cutout) then wtblock = wcut(wtblock)

   time_b = scan.time_s(wtblock) - scan.time_s(wtblock(0))
   time_s = [time_s, max(time_s) + timegap + time_b]
   tempo_subscans = scan.flag_scanleg(wtblock)
   tab_subscans = [tab_subscans, nt + tempo_subscans]
;;;
   if (not keyword_set(spiral)) then begin
      tab_sup = [vect_scanrot_deg(0)+dblarr(ntoff/2), vect_scanrot_deg, $
         vect_scanrot_deg(nt_b-1-ntoff)+dblarr(ntoff-ntoff/2)]
      tab_scanrot_deg = [tab_scanrot_deg, tab_sup(wtblock)]
   endif
;;;
   nt = nt + nt_b
   ind_scans = [ind_scans, nt]


   for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
      pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      Xra = (scan.(w(0)))(wtblock, 0:nbolo(detarr)-1)
      pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      Ydec = (scan.(w(0)))(wtblock, 0:nbolo(detarr)-1)

      if keyword_set(p_Xra) then begin
         p_Xra = Xra
         p_Ydec = Ydec
         gnomonic, p_Xra, p_Ydec, refx=p_mid_Xra, refy=p_mid_Ydec
         p_Xra *= 3600.D0
         p_Ydec *= 3600.D0
      endif

      gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
      Xra *= 3600.D0
      Ydec *= 3600.D0
      if keyword_set(offset_ra_as) then begin
         Xra -= offset_ra_as(i)
         Ydec += offset_dec_as(i)
      endif

      pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      *signals(detarr) = [*signals(detarr), float(scan.(w(0))(wtblock, 0:nbolo(detarr)-1))]
      elec_boxes(i, 0L:nbolo(detarr)-1L) = scan.mat_boxes - min(scan.mat_boxes, /nan)

      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         if keyword_set(p_Xra) then begin
            *p_xas_bolos(detarr) = [*p_xas_bolos(detarr), p_Xra * p_cos_scanrot + p_Ydec * p_sin_scanrot]
            *p_yas_bolos(detarr) = [*p_yas_bolos(detarr), p_Ydec * p_cos_scanrot - p_Xra * p_sin_scanrot]
         endif else begin
            *p_xas_bolos(detarr) = [*p_xas_bolos(detarr), Xra * p_cos_scanrot + Ydec * p_sin_scanrot]
            *p_yas_bolos(detarr) = [*p_yas_bolos(detarr), Ydec * p_cos_scanrot - Xra * p_sin_scanrot]
         endelse
      endif

      case orient of
         'astro': begin
            *xas_bolos(detarr) = [*xas_bolos(detarr), Xra]
            *yas_bolos(detarr) = [*yas_bolos(detarr), Ydec]
            end
         'scan': begin
            *xas_bolos(detarr) = [*xas_bolos(detarr), Xra * cos_scanrot + Ydec * sin_scanrot]
            *yas_bolos(detarr) = [*yas_bolos(detarr), Ydec * cos_scanrot - Xra * sin_scanrot]
            end
      endcase
   endif
;;;
   endif else begin
      nscans_block -= 1
      files_scans(i) = ''
   endelse
;;;
endfor
scan = 0

for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
      *p_xas_bolos(detarr) += (p_off_cen_Xra * 3600.D0)
      *p_yas_bolos(detarr) += (p_off_cen_Ydec * 3600.D0)
   endif
   *xas_bolos(detarr) += (off_cen_Xra * 3600.D0)
   *yas_bolos(detarr) += (off_cen_Ydec * 3600.D0)
endif

w = where(strlen(files_scans) gt 1)
elec_boxes = elec_boxes(w, *)
config = 1 + intarr(nscans_block)
for s = 1, nscans_block - 1 do begin
   w = where(max(elec_boxes(s, *), /nan) eq max(elec_boxes(0:s-1, *), dimension=2, /nan), cw)
   if cw ge 1 then config(s) = config(w(0)) else config(s) += 1
endfor
nconfig = max(config)
for conf = 2, nconfig do begin
   w = where(config eq conf)
   w0 = where(config eq conf - 1)
   elec_boxes(w, *) += (max(elec_boxes(w0, *), /nan) + 1)
endfor
if nconfig gt 1 then print, crs + 'dataset mixing ' + chain(nconfig) + ' different array geometries' + crs




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; save signal and coordinates to disk
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   signal = *signals(detarr)
   save, file=file_signal, signal, /xdr
   signal = 0
   file_xyas = root_xyas + arrays_default(detarr) + '.xdr'
   file_xyas_final = root_xyas + 'final_' + arrays_default(detarr) + '.xdr'
   xas = *xas_bolos(detarr)
   yas = *yas_bolos(detarr)
   save, file=file_xyas_final, xas, yas, /xdr
   if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
      xas = *p_xas_bolos(detarr)
      yas = *p_yas_bolos(detarr)
   endif
   save, file=file_xyas, xas, yas, /xdr
   xas = 0
   yas = 0
endif



;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; time-related quantities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; instantaneous scan speed:

w = where(ind_arrays_eff eq 1)
detarr = w(0)
vscan = fltarr(nt, 3)
for k_off = 0, 2 do begin
   off = k_off * 2L + 5L
   for i = 0, nscans_block - 1 do begin
      dd = sqrt(((*xas_bolos(detarr))(ind_scans(i)+off:ind_scans(i+1)-1, boloref(detarr)) $
         - (*xas_bolos(detarr))(ind_scans(i):ind_scans(i+1)-1-off, boloref(detarr)))^2. + $
         ((*yas_bolos(detarr))(ind_scans(i)+off:ind_scans(i+1)-1, boloref(detarr)) $
         - (*yas_bolos(detarr))(ind_scans(i):ind_scans(i+1)-1-off, boloref(detarr)))^2.)
      dt = time_s(ind_scans(i)+off:ind_scans(i+1)-1) - time_s(ind_scans(i):ind_scans(i+1)-1-off)
      vscan(ind_scans(i)+off/2+1:ind_scans(i+1)-1-off/2, k_off) = float(dd / dt)
   endfor
endfor
vscan = median(vscan, dimension=2)
dd = 0
dt = 0

vscan = smooth(vscan, 3)
w = where(vscan lt 0., cw)
if cw ge 50 then begin
   print, 'negative velocities !'
   if keyword_set(debug) then stop
   vscan(w) = 0.
endif
save, file=file_vscan, vscan, /xdr



;;; visualization of all bolometer trajectories:

if keyword_set(vis_traject) then begin
   w = where(ind_arrays_eff eq 1)
   detarr = w(0)
   window, 0, xs=800, ys=800
   min_xy = min([min(*xas_bolos(detarr), /nan), min(*yas_bolos(detarr), /nan)])
   max_xy = max([max(*xas_bolos(detarr), /nan), max(*yas_bolos(detarr), /nan)])
   if nblocks gt 1 then begin
      min_xy = 0.
      max_xy = (max([max_xdeg, max_ydeg]) + drdeg_max(detarr) / 2.) * 3600.
   endif
   if nbolo(detarr) gt 500 then ind_b = 4L * lindgen(nbolo(detarr)/4L) else ind_b = lindgen(nbolo(detarr))
   otherplot, (*xas_bolos(detarr))(0, ind_b), (*yas_bolos(detarr))(0, ind_b), psym=4, charsize=1.5, $
      xrange=[min_xy, max_xy], yrange=[min_xy, max_xy], xtitle='x coordinate (arcsec)', ytitle='y coordinate (arcsec)'
   if nt lt 100000L then dnt = 10L else dnt = 50L * min([1L, (1L + nt / 200000L)])
   for k = 0L, nscans_block - 1 do begin
      for i = 0L, (ind_scans(k+1) - ind_scans(k) - 1) / dnt do begin
         t = ind_scans(k) + dnt * i
         otherplot, /over, color=i*5, (*xas_bolos(detarr))(t, ind_b), (*yas_bolos(detarr))(t, ind_b), psym=4
      endfor
      wait, 0.5
   endfor
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   window, 0, xs=1200
endif



;;; median scan speed:

histo = histogram(vscan, locations=v, binsize=0.1, min=5.)
w = where(histo gt max(histo) / 2.)
v = v(w)
w = where(vscan ge 0.9 * min(v) and vscan le 1.1 * max(v))
vscan_as_s = median(vscan(w))


;;; check scan speed of each scan:
flag_speed = intarr(nscans)
for s = 0, nscans_block - 1 do begin
   v_s = median(vscan(ind_scans(s):ind_scans(s+1)-1))
   if abs(1. - v_s / vscan_as_s) gt 0.1 then flag_speed(s) = 1
endfor
vs_min = vscan_as_s
vs_max = vscan_as_s
if max(flag_speed) gt 0 then begin
   print, 'discordant scan speed (wrt the median) for:'
   for s = 0, nscans_block - 1 do if flag_speed(s) gt 0 then print, files_scans(s)
   w = where(flag_speed gt 0, cw)
   for k = 0, cw - 1 do begin
      v_s = median(vscan(ind_scans(w(k)):ind_scans(w(k)+1)-1))
      print, v_s
      vs_min = min([vs_min, v_s])
      vs_max = max([vs_max, v_s])
   endfor
   print, 'min and max scan speeds:'
   print, vs_min, vs_max
   vscan_as_s = (vs_min + vs_max) / 2.
endif


;;; flag vector for scan turnarounds (= 0 when non-zero acceleration, = 1 otherwise):

flag_nza = 1 + intarr(nt)
ntoff = 100
ntoff = min([ntoff, round(median(ind_scans(1:nscans_block) - ind_scans(0:nscans_block-1), /even)) / 20L])

if (not keyword_set(spiral)) then begin
   w = where(vscan lt 1. or vscan gt 3. * vscan_as_s, cw)
   if cw ge 1 then flag_nza(w) = 0

   tab_rot = tab_scanrot_deg
   w = where(tab_rot lt 0., cw)
   if cw ge 1 then tab_rot(w) += 360.D0
   w = where(tab_rot gt 360., cw)
   if cw ge 1 then tab_rot(w) -= 360.D0
   for s = 0, nscans_block - 1 do begin
      histo = histogram(tab_rot(ind_scans(s):ind_scans(s+1)-1), locations=ang, binsize=1., min=0.D0, max=360.D0)
      w = where(histo gt max(histo) / 4., cw)
      scanrot_1 = median(ang(w(0:cw/2-1)))
      scanrot_2 = median(ang(w(cw/2:cw-1)))
      if (scanrot_1 lt 10. or scanrot_2 gt 350.) then begin
         tab_rot(ind_scans(s):ind_scans(s+1)-1) += 90.D0
         w = where(tab_rot(ind_scans(s):ind_scans(s+1)-1) gt 360., cw)
         if cw ge 1 then tab_rot(ind_scans(s)+w) -= 360.D0
         histo = histogram(tab_rot(ind_scans(s):ind_scans(s+1)-1), locations=ang, binsize=1., min=0.D0, max=360.D0)
         w = where(histo gt max(histo) / 4., cw)
         scanrot_1 = median(ang(w(0:cw/2-1)))
         scanrot_2 = median(ang(w(cw/2:cw-1)))
      endif
      w = where(abs(tab_rot(ind_scans(s):ind_scans(s+1)-1) - scanrot_1) gt 20. $
         and abs(tab_rot(ind_scans(s):ind_scans(s+1)-1) - scanrot_2) gt 20., cw)
      if cw ge 1 then flag_nza(ind_scans(s)+w) = 0
   endfor
endif


if keyword_set(visu) then window, 0, xs=1200

if nblocks eq 1 then begin
   wup = where([0, flag_nza(1:*) - flag_nza(0:nt-2)] eq 1, cup)
   for i = 0L, cup - 1L do if wup(i) lt nt - 3L then begin
      wdown = where(flag_nza(wup(i):nt-1) - flag_nza(wup(i)-1:nt-2) eq (-1), cdown)
      if (cdown ge 1 and wdown(0) lt ntoff) then flag_nza(wup(i):wup(i)+wdown(0)) = 0
   endif
endif

save, file=file_times, time_s, flag_nza, /xdr



;;; nominal scan speed in arcsec/s and time step in s:

w = where(flag_nza eq 1, cw)
vscan_as_s = vs_max
print, 'nominal scan speed = ' + chain(vscan_as_s, 2) + ' arcsec/s'

diff_t = time_s(w(1:cw-1)) - time_s(w(0:cw-2))
w = 0
diff_t = diff_t(sort(diff_t))
timestep = float(median(diff_t(cw/10:9*cw/10-1)))
diff_t = 0



;;; subscan indices:

flag_drop = intarr(nt)

;;;
ind_subscans = [0L, 1L + where(tab_subscans(1L:*) - tab_subscans(0L:nt-2L) gt 0L), nt]
goto, skip_indices
;;;

ind_subscans = [0L]
for s = 0, nscans - 1 do begin
   it0 = ind_scans(s)
   repeat begin
      w = where(flag_nza(it0:ind_scans(s+1)-1) eq 1)
      it0 += max([w(0), 0])
      it1 = it0 + 1
      if it1 lt ind_scans(s+1) - 1 then w = where(flag_nza(it1:ind_scans(s+1)-1) eq 0 $
         or [time_s(it1+1:ind_scans(s+1)-1) - time_s(it1:ind_scans(s+1)-2), 0.] gt maxtimediff $
         or [abs(tab_rot(it1+1:ind_scans(s+1)-1) - tab_rot(it1:ind_scans(s+1)-2)), 0.] gt 90., cw) $
       else cw = 0
      if cw eq 0 then it1 = ind_scans(s+1) - 1 else begin
         it1 += w(0)
         if flag_nza(it1) eq 0 then begin
            w = where(flag_nza(it1:ind_scans(s+1)-1) eq 1, cw)
            if cw eq 0 then it1 = ind_scans(s+1) - 1 else it1 += (w(0) / 2)
         endif
      endelse
      if (it1 ge ind_scans(s+1) - 31 and it1 lt ind_scans(s+1) - 1) then begin
         flag_drop(it1+1:ind_scans(s+1)-1) = 1
         it1 = ind_scans(s+1) - 1
      endif
      it0 = it1 + 1
      ind_subscans = [ind_subscans, it0]
   endrep until it0 ge ind_scans(s+1)
endfor

;;;
skip_indices:
;;;

nsubscans = n_elements(ind_subscans) - 1L



;;; remove very short scan legs:

flag_sup = intarr(nsubscans+1)
w = where(ind_arrays_eff eq 1)
detarr = w(0)
s = -1
for s_tot = 0, nscans - 1 do if strlen(files_scans(s_tot)) gt 1 then begin
   s += 1
   if nblocks eq 1 then begin
      restore, dir_in + files_scans(s_tot)
      winit = where(scan.flag_scanleg gt 0L)
   endif
   wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1, csubs)
   print, 'scan ' + chain(s) + ' : subscans ' + chain(wsubs(0)) + ' to ' + chain(wsubs(csubs-1))
   for subs = wsubs(0), wsubs(csubs-1) do begin
      nb_subscan = total(finite((*signals(detarr))(ind_subscans(subs):ind_subscans(subs+1)-1, *)), 2)
      w = where(nb_subscan gt nbolo(detarr) / 10L, cw)
      if ((cw lt (ind_subscans(subs+1) - ind_subscans(subs)) / 2L and nblocks eq 1) $
       or ind_subscans(subs+1) - ind_subscans(subs) lt 100L) then begin
         flag_sup(subs) = 1
         print, '     subscan ' + chain(subs) + ' has no or little valid data'
         if nblocks eq 1 then scan.flag_scanleg(winit(ind_subscans(subs)-ind_scans(s):ind_subscans(subs+1)-1-ind_scans(s))) = $
            -abs(scan.flag_scanleg(winit(ind_subscans(subs)-ind_scans(s):ind_subscans(subs+1)-1-ind_scans(s))))
      endif
   endfor
   if nblocks eq 1 then save, file=dir_in+files_scans(s_tot), scan, /xdr
endif
scan = 0

;;;
nscans = nscans_block
;;;


w = where(flag_sup gt 0, cw, complement=wgood, ncomplement=cgood)
if cw ge 1 then begin
   print, chain(cw) + ' removed subscan(s)'
;;;
for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
;;;
   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   signal = *signals(detarr)
   file_xyas = root_xyas + arrays_default(detarr) + '.xdr'
   file_xyas_final = root_xyas + 'final_' + arrays_default(detarr) + '.xdr'
   xas = *xas_bolos(detarr)
   yas = *yas_bolos(detarr)
   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      xas_p = *p_xas_bolos(detarr)
      yas_p = *p_yas_bolos(detarr)
   endif

   if w(0) eq 0 then begin
      ind_next = ind_subscans(wgood(0))
      signal = signal(ind_next:*, *)
      xas = xas(ind_next:*, *)
      yas = yas(ind_next:*, *)
      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         xas_p = xas_p(ind_next:*, *)
         yas_p = yas_p(ind_next:*, *)
      endif
      vscan = vscan(ind_next:*)
      flag_nza = flag_nza(ind_next:*)
      time_s = time_s(ind_next:*)
      csubs = ind_next - ind_subscans(0)
      ind_scans(1:*) -= csubs
      ind_subscans(1:*) -= csubs
   endif

   for k = 0L, cw - 1L do if (w(k) gt wgood(0) and w(k) lt wgood(cgood-1) and w(k) lt nsubscans - 1L) then begin
      subs = w(k)
      signal = [signal(0L:ind_subscans(subs)-1L, *), signal(ind_subscans(subs+1):*, *)]
      xas = [xas(0L:ind_subscans(subs)-1L, *), xas(ind_subscans(subs+1):*, *)]
      yas = [yas(0L:ind_subscans(subs)-1L, *), yas(ind_subscans(subs+1):*, *)]
      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         xas_p = [xas_p(0L:ind_subscans(subs)-1L, *), xas_p(ind_subscans(subs+1):*, *)]
         yas_p = [yas_p(0L:ind_subscans(subs)-1L, *), yas_p(ind_subscans(subs+1):*, *)]
      endif
      vscan = [vscan(0L:ind_subscans(subs)-1L), vscan(ind_subscans(subs+1):*)]
      flag_nza = [flag_nza(0L:ind_subscans(subs)-1L), flag_nza(ind_subscans(subs+1):*)]
      time_s = [time_s(0L:ind_subscans(subs)-1L), time_s(ind_subscans(subs+1):*)]
      csubs = ind_subscans(subs+1) - ind_subscans(subs)
      ws = where(ind_scans ge ind_subscans(subs+1), cws)
      if cws ge 1 then ind_scans(ws) -= csubs
      ind_subscans(subs+1:*) -= csubs
   endif

   if w(cw-1L) eq nsubscans - 1L then begin
      subs = wgood(cgood-2) + 1
      signal = signal(0L:ind_subscans(subs)-1L, *)
      xas = xas(0L:ind_subscans(subs)-1L, *)
      yas = yas(0L:ind_subscans(subs)-1L, *)
      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         xas_p = xas_p(0L:ind_subscans(subs)-1L, *)
         yas_p = yas_p(0L:ind_subscans(subs)-1L, *)
      endif
      vscan = vscan(0L:ind_subscans(subs)-1L)
      flag_nza = flag_nza(0L:ind_subscans(subs)-1L)
      time_s = time_s(0L:ind_subscans(subs)-1L)
      csubs = ind_subscans(subs+1) - ind_subscans(subs)
      ind_scans(nscans) -= csubs
      ind_subscans(subs+1) -= csubs
   endif

   ind_subscans = ind_subscans(wgood)
   nt = n_elements(time_s)

   save, file=file_vscan, vscan, /xdr
   save, file=file_times, time_s, flag_nza, /xdr
   save, file=file_signal, signal, /xdr
   *signals(detarr) = signal

   save, file=file_xyas_final, xas, yas, /xdr
   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      xas = xas_p
      yas = yas_p
   endif
   save, file=file_xyas, xas, yas, /xdr
endif
;;;
endif
;;;


nsubscans = n_elements(ind_subscans) - 1
;;;
ptr_free, xas_bolos, yas_bolos
if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then ptr_free, p_xas_bolos, p_yas_bolos
;;;



if keyword_set(visu) then begin
   wset, 0
   tab_nsubs = ind_subscans(1:nsubscans) - ind_subscans(0:nsubscans-1)
   otherplot, time_s(ind_subscans), tab_nsubs, psym=-4, xtitle='time (s)', ytitle='N', $
      title='time samples per scan leg (must be regular in each scan)', yrange=[0., 1.1*max(tab_nsubs)]
   for scan = 0, nscans - 1 do otherplot, /over, time_s(ind_scans(scan))+[0.,0.], [0., 1.1*max(tab_nsubs)], color=30
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
endif



;;; save signal again after masking very low or very high speeds:

for s = 0L, nscans - 1L do begin
   wnom = where(flag_nza(ind_scans(s):ind_scans(s+1)-1) eq 1)
   wini = max(wnom) + 1
   if ind_scans(s) + wini lt ind_scans(s+1) - 10 then begin
      do_stop = 0
      repeat begin
         wstart = where(vscan(ind_scans(s)+wini:ind_scans(s+1)-1) lt 0.1, cw)
         if cw ge 1 then begin
            wstart = wini + wstart(0)
            wend = where(vscan(ind_scans(s)+wstart:ind_scans(s+1)-1) ge 0.1, ce)
            if (ce lt 1 or wend(0) ge 100) then begin
               flag_drop(ind_scans(s)+wstart:ind_scans(s+1)-1) = 1
               do_stop = 1
            endif else wini = wstart + wend(0)
         endif else do_stop = 1
      endrep until do_stop eq 1
   endif
endfor

w = where(vscan lt 0.1 or vscan gt 3. * vscan_as_s or flag_drop gt 0, cw)
if keyword_set(spiral) then w = where(vscan lt 0.1 or vscan gt 10. * vscan_as_s or flag_drop gt 0, cw)
if nzadata eq 'no' then w = where(flag_nza lt 1 or vscan lt vscan_as_s / 2. or vscan gt 1.5 * vscan_as_s, cw)

if cw ge 1 then for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   signal = *signals(detarr)
   signal(w, *) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0
endif

ptr_free, signals



if keyword_set(visu) then begin
   wset, 0
   scan = 0
   otherplot, time_s(ind_scans(scan):ind_scans(scan+1)-1), vscan(ind_scans(scan):ind_scans(scan+1)-1), $
      xtitle='t (s)', ytitle='V!lscan!n (arcsec/s)', $
      title='scan speed of scan '+chain(scan), yrange=[0.8*vs_min, 1.25*vs_max]
   otherplot, /over, time_s(ind_scans(scan):ind_scans(scan+1)-1), $
      flag_nza(ind_scans(scan):ind_scans(scan+1)-1) * 1.2*vscan_as_s, color=200.
   if nsubscans le 40 then for subs = 0L, nsubscans - 1L do $
      otherplot, /over, time_s(ind_subscans(subs))+[0.,0.], [0., 1.25*vscan_as_s], color=30, lstyle=2.
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   otherplot, time_s, vscan, xtitle='t (s)', ytitle='V!lscan!n (arcsec/s)', $
      title='scan speed of all scans', yrange=[0., 2.*vs_max]
   if nsubscans le 40 then otherplot, /over, time_s, flag_nza * 1.5*vscan_as_s, color=200.
   for scan = 0, nscans - 1 do otherplot, /over, time_s(ind_scans(scan))+[0.,0.], $
      [0., 2.*vscan_as_s], color=30
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
endif


end
