function subtract_aver_drift_artemis, drifts_all, flag_valid_scans=flag_valid_scans, maxnoise=maxnoise, $
         highfnoise=highfnoise, ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, $
         ind_aver=ind_aver, file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
         t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, iter=iter, galactic=galactic, $
         size_deg=size_deg, dr_bolo=dr_bolo, time_1=time_1, t_proc=t_proc


;;; computes average drift, interpolates and optionally subtracts it from the signal

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

crs = string(13b) + string(10b)


;;;
separate_matrices = 0
;;;separate_matrices = 1
;;;


b_refpos = 29L + 64L * 15L
nmat = 1

;;;
if keyword_set(separate_matrices) then begin
restore, file_signal
signal_orig = signal
signal = 0
file_signal_orig = strmid(file_signal, 0, strlen(file_signal)-4) + '_orig.xdr'
save, file=file_signal_orig, signal_orig, /xdr
signal_orig = 0
restore, file_maps
map_signal_orig = map_signal
map_weight_orig = map_weight
maxnoise_orig = maxnoise
highfnoise_orig = highfnoise

nb_row = 16L
nb_col = 18L
nb_mat = nb_row * nb_col
nmat = nb / nb_mat
if nmat eq 8L then nsub_x = 4L else nsub_x = 2L
nb_line = nsub_x * nb_row

amp_drifts_all = fltarr(nsubscans, nb)
endif
;;;


;;;
for mat = 0, nmat - 1 do begin
   if keyword_set(separate_matrices) then begin
      wmat = reform(rebin((mat mod nsub_x) * nb_row + lindgen(nb_row), nb_row, nb_col) $
         + nb_line * (((mat / nsub_x) * nb_col + lindgen(nb_col)) ## (1L + lonarr(nb_row))), nb_mat)
      restore, file_signal_orig
      signal = !values.f_nan + fltarr(nt, nb)
      signal(*, wmat) = signal_orig(*, wmat)
      signal_orig = 0
      save, file=file_signal, signal, /xdr

      maxnoise = !values.f_nan + fltarr(nsubscans, nb)
      maxnoise(*, wmat) = maxnoise_orig(*, wmat)
      highfnoise(*, wmat) = highfnoise_orig(*, wmat)

      prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
      prepare_proj_coarse, file_scalars=file_scalars
      proj_map_artemis, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, $
         ind_subscans=ind_subscans
   endif
;;;



compute_aver_drift_artemis, drifts_save, npairs_save, w_glitches, nb_glitch, maxnoise=maxnoise, highfnoise=highfnoise, $
   ind_scans=ind_scans, ind_aver=ind_aver, file_scalars=file_scalars, min_nsamp=min_nsamp, $
   detect_gl=detect_gl, t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, galactic=galactic, $
   time_1=time_1, t_proc=t_proc, ind_subscans=ind_subscans


;;; mask glitches and update projection vectors:
if nb_glitch ge 1 then begin
   print, crs + 'removed ' + chain(nb_glitch) + ' sample(s) affected by glitches'
   restore, file_signal
   signal(w_glitches) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0
   w_glitches = 0

   prepare_proj_coarse, file_scalars=file_scalars

   if (keyword_set(visu) and nb_glitch gt 100) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
      prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
      proj_map_nearest_artemis, map_noglitch, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans
      disp_ima, map_signal - map_noglitch, win=4, title='map of residual glitches', $
         weightmap=map_weight, /trunc_max, noise_level=median(maxnoise)
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 4
      map_noglitch = 0
      map_signal = 0
      map_weigth = 0
      time_1 = systime(/seconds)
   endif
endif



restore, file_times

restore, file_time_coarse


drifts = fltarr(nt)
amp_drifts = fltarr(nsubscans, nb)

npairs_min = 0.


;;; cancel correction if it is too sparsely determined:

flag_nza_coarse = interpol(flag_nza, float(time_s), time_save)
wnom = where(flag_nza_coarse gt 0., cnom)
wdef = where(flag_nza_coarse gt 0. and npairs_save gt npairs_min, cdef)
ratio_def = float(cdef) / float(cnom)
;;;
if keyword_set(debug) then begin
   print, 'ratio_def = ' + chain(ratio_def, 3)
endif
;;;
if ratio_def lt 0.7 then begin
   print, '              average drift too sparsely determined -> correction cancelled.'
   amp_drifts = fltarr(nsubscans, nb)
   return, amp_drifts
endif


;;; interpolate on fine time grid :

;;;
restore, file_signal
wb = where(total(finite(maxnoise), 1) ge nsubscans / 4L, nb_valid)
wz = where(total(finite(signal), 2) lt nb_valid / 4L, cz, complement=wdef, ncomplement=cdef)
sig_med = smooth(median(signal, dimension=2, /even), 7L, /nan)
if cz ge 1 then sig_med(wz) = !values.f_nan
signal = 0
sig_med = interpol(sig_med(wdef), float(time_s(wdef)), time_save)

restore, file_vscan
vscan_coarse = interpol(vscan, float(time_s), time_save)
vscan = 0
;;;

wz = where(npairs_save le npairs_min, cz)
if cz ge 1 then drifts_save(wz) = !values.f_nan

;;;npairs_interp = fltarr(nt)
for i = 0, nscans - 1 do begin
   time_ref = time_s(ind_scans(i))
   csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
   wsave = ind_sc_coarse(i) + lindgen(csave)
   wf = where(finite(drifts_save(wsave)) gt 0)
   wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1, csubs)
;;;   wz = where(finite(drifts_save(wsave)) lt 1, cz, complement=wdef, ncomplement=cdef)
   wz = where(finite(drifts_save(wsave)) lt 1 $
      or vscan_coarse(wsave) le 1. or vscan_coarse(wsave) ge 2. * vscan_as_s, cz, complement=wdef, ncomplement=cdef)
;;;

   if cdef ge 15 then begin
      if keyword_set(debug) then begin
         wset, 0
         otherplot, time_save(wsave(wf)), drifts_save(wsave(wf)), psym=-4, $
            xtitle='t (s) (coarse grid)', ytitle='(true + excess) drift (Jy/bol.)', $
            title='scan ' + chain(i+1) +  ': original drift in black, interpolated drift in red'
      endif

      if cz ge 1 then begin
         tab_drifts = drifts_save(wsave)
         tab_drifts(wz) = !values.f_nan
;;;         if ratio_def lt 0.95 then wdef = where(smooth(float(finite(tab_drifts)), 15) ge 0.7 and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
;;;         wdef = where((smooth(float(finite(tab_drifts)), 15) ge 0.7 or flag_nza_coarse lt 1.) and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
         wdef = where(finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
;;;
         if cdef ge 11 then begin
            tab_drifts(wz) = !values.f_nan
            if cdef ge 3 then begin

;;;
;;; median signal interpolation for scanleg ends:
;;;
if wdef(0) gt 0 then tab_drifts(0L:wdef(0)-1L) = tab_drifts(wdef(0))
if wdef(cdef-1L) lt csave - 1L then tab_drifts(wdef(cdef-1L)+1:csave-1L) = tab_drifts(wdef(cdef-1L))
for subs = wsubs(0), wsubs(csubs-2) do begin
   w1 = where(time_save(wsave) le time_s(ind_subscans(subs+1)-1L) and finite(tab_drifts) gt 0, c1)
   w1 = w1(c1-1)
   w2 = where(time_save(wsave) ge time_s(ind_subscans(subs+1)) and finite(tab_drifts) gt 0)
   w2 = w2(0)
   if w2 gt w1 + 1 then tab_drifts(w1+1:w2-1) = sig_med(wsave(w1+1:w2-1)) - mean(sig_med(wsave([w1, w2]))) $
      + mean(tab_drifts([w1, w2]))
endfor
wdef = where(finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
;;;


               for k = 0, cz - 1 do begin
                  m = min(abs(wz(k) - wdef), ind)
                  flag_end = 0
                  flag_beg = 0
                  if wdef(ind) lt wz(k) then begin
                     ind_1 = ind
                     ind_2 = ind + 1
                     if ind_2 ge cdef then begin
                        ind_1 = ind - 3
                        ind_2 = ind - 2
                        flag_end = 1
                     endif
                  endif else begin
                     ind_1 = ind - 1
                     ind_2 = ind
                     if ind_1 lt 0 then begin
                        ind_1 = ind + 2
                        ind_2 = ind + 3
                        flag_beg = 1
                     endif
                  endelse
                  di = wdef(ind_2) - wdef(ind_1)
                  tabx1 = wdef(ind_1) - reverse(lindgen(min([di+1, wdef(ind_1)+1])))
                  tabx2 = wdef(ind_2) + lindgen(min([di+1, csave-wdef(ind_2)]))
                  taby1 = tab_drifts(tabx1)
                  taby2 = tab_drifts(tabx2)
                  x = [total(tabx1 * finite(taby1)) / total(finite(taby1)), total(tabx2 * finite(taby2)) / total(finite(taby2))]
                  y = [total(taby1, /nan) / total(finite(taby1)), total(taby2, /nan) / total(finite(taby2))]
                  if flag_end eq 1 then y(0) = y(1)
                  if flag_beg eq 1 then y(1) = y(0)
                  tab_drifts(wz(k)) = interpol(y, x, wz(k))
               endfor
               drifts_save(wsave) = tab_drifts
            endif else drifts_save(wsave(wz)) = 0.
            npairs_save(wsave(wz)) = 0.
         endif else begin
            drifts_save(wsave) = 0.
            npairs_save(wsave) = 0.
         endelse
      endif

      if keyword_set(debug) then begin
         otherplot, /over, color=10, psym=-4, time_save(wsave), drifts_save(wsave)
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif

      tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
      tab_drifts = [drifts_save(wsave(0)), drifts_save(wsave), drifts_save(wsave(csave-1))]
      subs_drifts = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
      drifts(ind_scans(i):ind_scans(i+1)-1) = subs_drifts
;;;      npairs_interp(ind_scans(i):ind_scans(i+1)-1) = interpol(npairs_save(wsave), $
;;;         -float(time_ref) + time_save(wsave), float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
   endif else begin
      drifts_save(wsave) = 0.
      npairs_save(wsave) = 0.
   endelse
endfor



;;; remove drift common to all cross-scans
;;; (solution to the problem of deriving the absolute drift from the drift differences matrix
;;; is not unique, and is affected by oscillations with a periodicity of 2 scans, i.e. 1 cross-scan,
;;; that can be removed in this way):

restore, file_signal
signal_save = signal
restore, file_maps
map_signal_save = map_signal
map_weight_save = map_weight
if (not keyword_set(map_signal_uncl)) then map_signal_uncl = 0
map_uncl_save = map_signal_uncl

restore, file_pixcoords



;;;
drifts_save_init = drifts_save
goto, skip_excess
;;;


   map_drift = fltarr(dimx, dimy, nscans)
   weight_drift = fltarr(dimx, dimy, nscans)
;;;   if size_deg lt 1. then ns_lim = 30 else ns_lim = 15
ns_lim = nscans + 10
;;;

;;;
if (not keyword_set(separate_matrices)) then signal = rebin(drifts, nt, nb) else begin
   signal = !values.f_nan + fltarr(nt, nb)
   signal(*, wmat) = rebin(drifts, nt, nb_mat)
endelse
;;;
   w = where(npairs_interp le npairs_min, cw)
   if cw ge 1 then signal(w, *) = !values.f_nan
;;;
if (not keyword_set(separate_matrices)) then simnoise = 1. + fltarr(nsubscans, nb) else begin
   simnoise = !values.f_nan + fltarr(nsubscans, nb)
   simnoise(*, wmat) = 1.
endelse
;;;

   ;;; keep only the central part of the array to project the drift,
   ;;; in order to suppress the noise due to coverage non-uniformities and scan leg edges:
   if keyword_set(dr_bolo) then for i = 0, nscans - 1 do begin
      it_mid = (ind_scans(i)+ind_scans(i+1)-1) / 2
;;;
if (not keyword_set(separate_matrices)) then begin
   dx = abs(reform(x_bolos(it_mid, *)) - x_bolos(it_mid, b_refpos))
   dy = abs(reform(y_bolos(it_mid, *)) - y_bolos(it_mid, b_refpos))
endif else begin
      dx = abs(reform(x_bolos(it_mid, *)) - median(x_bolos(it_mid, wmat), /even))
      dy = abs(reform(y_bolos(it_mid, *)) - median(y_bolos(it_mid, wmat), /even))
endelse
;;;
      r = sqrt(dx^2. + dy^2.)
      w = where(r gt dr_bolo)
      signal(ind_scans(i):ind_scans(i+1)-1, w) = !values.f_nan
      wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1, csubs)
      for isubs = 0L, csubs - 1L do simnoise(wsubs(isubs), w) = !values.f_nan
      w = where(r le dr_bolo)
      for isubs = 0L, csubs - 1L do simnoise(wsubs(isubs), w) *= (1. + r(w) / dr_bolo)
   endfor

   save, file=file_signal, signal, /xdr

   cut_pcal = 0
   do_map_scans_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, maxnoise=simnoise
   map_drift = map_scans
   map_scans = 0
   weight_drift = weight_scans
   weight_scans = 0


   w = where(weight_drift gt 0.)
   med_weight = median(weight_drift(w))
   sig_weight = sqrt(mean((weight_drift(w) - med_weight)^2.))
   w = where(weight_drift gt 0. and abs(weight_drift - med_weight) lt 1.5 * sig_weight)
   sig_weight = sqrt(mean((weight_drift(w) - med_weight)^2.))
   thres_weight = max([med_weight - 2.5 * sig_weight, 0.01 * med_weight])
   w = where(weight_drift lt thres_weight)
   map_drift(w) = !values.f_nan
   weight_drift(w) = 0.


   w_sm = 3
   r_crit = 12

   if nscans lt ns_lim then begin
      if sqrt(dimx * dimy) gt 300 then width = 51L else width = 31L
      tab_width = [width, width, 1L]
      map_mask = map_weight_save
      w = where(map_weight_save gt 0.)
      med_weight = median(map_weight_save(w))
      sig_weight = sqrt(mean((map_weight_save(w) - med_weight)^2.))
      w = where(map_weight_save gt 0. and abs(map_weight_save - med_weight) lt 1.5 * sig_weight)
      sig_weight = sqrt(mean((map_weight_save(w) - med_weight)^2.))
      th_w = max([med_weight - 2.5 * sig_weight, 0.01 * med_weight])
      w = where(map_mask lt th_w and total(weight_drift, 3) le 0)
      map_mask(w) = !values.f_nan
      wall = where(map_weight_save gt 0.)
      length_all = float(1 + max(wall mod dimx) - min(wall mod dimx))
      length_all *= floor(dimx / length_all)
      for i = 0, nscans - 1 do begin
         wi = where(weight_drift(*, *, i) gt 0.)
         length = 1 + max(wi mod dimx) - min(wi mod dimx)
         if length lt 0.75 * length_all then length = round(dimx / length_all * length) / 4 else length = dimx / 4
         map_drift(*, *, i) = interpol_map(map_drift(*, *, i), dimx=dimx, dimy=dimy, $
            r_crit=r_crit, map_mask=map_mask, length=length)
      endfor

      map_sq = map_drift^2.
      ;;; do NOT use the /edge_truncate keyword in smooth
      map_smooth = smooth(map_drift, tab_width, /nan)
      map_smoothsq = smooth(map_sq, tab_width, /nan)
      map_dev2 = map_smoothsq - map_smooth^2.
      w = where(map_dev2 le 0. or finite(map_drift) lt 1, cw)
      if cw ge 1 then map_dev2(w) = !values.f_nan

      map_ref = total(map_drift / map_dev2, 3, /nan) / total(1. / map_dev2, 3, /nan)
      if nscans eq 3 then w = where(total(weight_drift gt 0., 3) le 2) else w = where(total(weight_drift gt 0., 3) le 1)
      map_ref(w) = !values.f_nan
      map_ref = smooth(map_ref, w_sm, /nan)
      map_ref(w) = !values.f_nan
      w = where(weight_drift le 0.)
      map_drift(w) = !values.f_nan
   endif


;;;   map_excess = median(map_drift, dimension=3, /even)
map_excess = total(map_drift, 3, /nan) / total(finite(map_drift), 3)
;;;if (not keyword_set(disjoint)) then map_excess = median(map_drift, dimension=3, /even)
;;;
   weight_excess = total(weight_drift, 3)
   wz_excess = where(finite(map_excess) lt 1)
;;;
   if keyword_set(disjoint) then begin
      map_cov = total(smooth(weight_drift, [21, 21, 1]) gt 0.1, 3)
      w = where(finite(map_excess) gt 0 and map_cov gt 2.5, cw)
      if cw ge 1 then map_excess(w) = 0.
   endif
;;;
   map_excess = smooth(map_excess, w_sm, /nan)
   map_excess(wz_excess) = !values.f_nan
;;;
   if (nscans lt ns_lim and (not keyword_set(disjoint))) then begin
;;;
      map_ref = interpol_map(map_ref, dimx=dimx, dimy=dimy, r_crit=r_crit, wmask=wz_excess, /force_x)
      w = where(finite(map_ref) gt 0)
      map_excess(w) = map_ref(w)
   endif
;;;
map_excess = median(map_ref, 21)
w = where(finite(map_ref) gt 0)
map_excess(w) = map_ref(w)
map_excess = smooth(map_excess, [21, 21], /nan)
map_excess(w) = map_ref(w)
stop
;;;


   if keyword_set(debug) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      w = where(finite(map_drift) gt 0, cw)
      ind_sort = sort(map_drift(w))
      min_d = map_drift(w(ind_sort(100*nscans)))
      max_d = map_drift(w(ind_sort(cw-1-100*nscans)))
      if max_d - min_d gt 10. * median(maxnoise) then pow_int=0.5 else pow_int = 1.
      disp_ima, win=10, map_drift(*, *, 0), weightmap=weight_drift(*, *, 0), $
         title='mapping of scan 0 average drift', min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      disp_ima, win=11, map_drift(*, *, 1), weightmap=weight_drift(*, *, 1), $
         title='mapping of scan 1 average drift', min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      if nscans lt ns_lim then begin
         weight_ref = fltarr(dimx, dimy)
         w = where(finite(map_ref) gt 0)
         weight_ref(w) = 10.
         disp_ima, win=9, map_ref, weightmap=weight_ref, title='reference drift map', $
            min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      endif else disp_ima, win=9, map_excess, weightmap=weight_excess, title='excess drift map', $
            min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      wdelete, 9, 10, 11
      time_1 = systime(/seconds)
   endif


;;;
if (not keyword_set(separate_matrices)) then $
   indxy_bref = floor(interpol(x_bolos(*, b_refpos), float(time_s), time_save)) $
      + dimx * floor(interpol(y_bolos(*, b_refpos), float(time_s), time_save)) $
   else indxy_bref = floor(interpol(median(x_bolos(*, wmat), dimension=2, /even), float(time_s), time_save)) $
      + dimx * floor(interpol(median(y_bolos(*, wmat), dimension=2, /even), float(time_s), time_save))
;;;
   x_bolos = 0
   y_bolos = 0
   simul_excess = map_excess(indxy_bref)


   drifts_save_init = drifts_save
   drifts_save = drifts_save_init - simul_excess


;;;
skip_excess:


   if keyword_set(debug) then begin
      wset, 0
      w = where(npairs_save gt npairs_min and finite(drifts_save) gt 0.)
      otherplot, time_save(w), drifts_save_init(w), xtitle='t (s) (coarse grid)', $
         ytitle='average drift (Jy/bol.)', title='(true + excess) drift in black, excess drift in yellow'
      if keyword_set(simul_excess) then otherplot, /over, color=200, time_save(w), simul_excess(w)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif


;;;
goto, skip_interp_bis
;;;
   for i = 0L, nscans - 1 do begin
      time_ref = time_s(ind_scans(i))
      csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
      wsave = ind_sc_coarse(i) + lindgen(csave)
      wz = where(npairs_save(wsave) le npairs_min or finite(drifts_save(wsave)) lt 1, cz, $
         complement=wdef, ncomplement=cdef)
      if cdef ge 15 then begin
         if cz ge 1 then begin
            tab_drifts = drifts_save(wsave)
;;;            if ratio_def lt 0.95 then wdef = where(npairs_save(wsave) gt npairs_min and smooth(float(finite(tab_drifts)), 15) ge 0.7 and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
;;;            wdef = where(npairs_save(wsave) gt npairs_min and (smooth(float(finite(tab_drifts)), 15) ge 0.7 or flag_nza_coarse lt 1.) and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
            wdef = where(npairs_save(wsave) gt npairs_min and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
;;;
            if cdef ge 3 then begin
               tab_drifts(wz) = !values.f_nan
               for k = 0, cz - 1 do begin
                  m = min(abs(wz(k) - wdef), ind)
                  flag_end = 0
                  flag_beg = 0
                  if wdef(ind) lt wz(k) then begin
                     ind_1 = ind
                     ind_2 = ind + 1
                     if ind_2 ge cdef then begin
                        ind_1 = ind - 3
                        ind_2 = ind - 2
                        flag_end = 1
                     endif
                  endif else begin
                     ind_1 = ind - 1
                     ind_2 = ind
                     if ind_1 lt 0 then begin
                        ind_1 = ind + 2
                        ind_2 = ind + 3
                        flag_beg = 1
                     endif
                  endelse
                  di = wdef(ind_2) - wdef(ind_1)
                  tabx1 = wdef(ind_1) - reverse(lindgen(min([di+1, wdef(ind_1)+1])))
                  tabx2 = wdef(ind_2) + lindgen(min([di+1, csave-wdef(ind_2)]))
                  taby1 = tab_drifts(tabx1)
                  taby2 = tab_drifts(tabx2)
                  x = [total(tabx1 * finite(taby1)) / total(finite(taby1)), total(tabx2 * finite(taby2)) / total(finite(taby2))]
                  y = [median(taby1, /even), median(taby2, /even)]
                  dev = [sqrt(total((taby1 - y(0))^2., /nan) / total(finite(taby1))), sqrt(total((taby2 - y(1))^2., /nan) / total(finite(taby2)))]
                  w = where(abs(y) lt 0.5 * dev, cw)
                  if cw ge 1 then y(w) = 0.
                  if flag_end eq 1 then y(0) = y(1)
                  if flag_beg eq 1 then y(1) = y(0)
                  tab_drifts(wz(k)) = interpol(y, x, wz(k))
               endfor
               drifts_save(wsave) = tab_drifts
            endif else drifts_save(wsave) = 0.
         endif
         tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
         tab_drifts = [drifts_save(wsave(0)), drifts_save(wsave), drifts_save(wsave(csave-1))]
         drifts(ind_scans(i)) = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
      endif else begin
         drifts_save(wsave) = 0.
         drifts(ind_scans(i):ind_scans(i+1)-1) = 0.
      endelse
   endfor
;;;
skip_interp_bis:
;;;




;;; remove zero level of drifts:
off = median(drifts)
drifts -= off
drifts_save -= off


signal = signal_save
save, file=file_signal, signal, /xdr
map_signal = map_signal_save
map_weight = map_weight_save
map_signal_uncl = map_uncl_save
save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, map_signal_uncl, /xdr
prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans


w = where(npairs_save gt npairs_min)
med = median(drifts_save(w))
for i = 0L, nsubscans - 1L do begin
   w = where(flag_nza(ind_subscans(i):ind_subscans(i+1)-1) gt 0)
   amp_drifts(i, *) = 3. * sqrt(mean((drifts(ind_subscans(i)+w) - med)^2.))
endfor



;;; visualize result:

if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   window, 0, xs=1200
   w = where(finite(signal) lt 1, cw)
   mask_bolo = 1 + intarr(nt, nb)
   if cw ge 1 then mask_bolo(w) = 0
   signal_aver = total(signal * mask_bolo, 2, /nan)
   norm = total(mask_bolo, 2)
   mask_bolo = 0
;;;
if (not keyword_set(separate_matrices)) then $
   wdef = where(norm gt nb / 5, complement=wz, ncomplement=nz) $
 else wdef = where(norm gt nb_mat / 5, complement=wz, ncomplement=nz)
;;;
   signal_aver(wdef) /= norm(wdef)
   if nz ge 1 then signal_aver(wz) = !values.f_nan
   med = median(signal_aver(wdef))
   dev = sqrt(mean((signal_aver(wdef) - med)^2.))
   for k = 0, 2 do begin
;;;
if (not keyword_set(separate_matrices)) then $
   wdev = where(norm gt nb / 5. and abs(signal_aver - med) lt 6. * dev) $
 else wdev = where(norm gt nb_mat / 5. and abs(signal_aver - med) lt 6. * dev)
;;;
      med = median(signal_aver(wdev))
      dev = sqrt(mean((signal_aver(wdev) - med)^2.))
   endfor
   dev = max([dev, median(maxnoise)])
   norm = 0
   yrange = [-4.*dev, 4.*dev]
   otherplot, time_s(wdef), signal_aver(wdef) - med, charsize=1.5, yrange=yrange, $
      title=title, xtitle='time (s)', ytitle='brightness + offset'
   w = where(npairs_save gt 0)
   otherplot, /over, time_save(w), drifts_save(w) - 2. * dev, psym=4, color=10
   otherplot, /over, time_s, drifts - 2. * dev, color=100
   for i = 0, nscans - 2 do begin
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)-1)+fltarr(2), yrange
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1))+fltarr(2), yrange
   endfor
   if nsubscans le 30 then begin
      otherplot, /over, time_s, dev * (13.8 * flag_nza - 10.), color=200
      for subs = 0L, nsubscans - 1 do otherplot, /over, time_s(ind_subscans(subs))+[0.,0.], yrange, color=30, lstyle=2.
   endif
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

   if keyword_set(simul_excess) then begin
      title = 'zoom: raw drift in grey, excess drift in yellow, corrected average drift below'
      yrange = [-4.*dev, 10.*dev]
   endif else title = 'zoom: corrected average drift below'
   otherplot, time_s(wdef), signal_aver(wdef) - med, charsize=1.5, yrange=yrange, $
      title=title, xtitle='time (s)', ytitle='brightness + offset', $
      xrange=time_s([ind_subscans(0), ind_subscans(min([4, nsubscans]))-1])
   w = where(npairs_save gt npairs_min)
   otherplot, /over, time_save(w), drifts_save(w) - dev, psym=4, color=10
   otherplot, /over, time_s, drifts - dev, color=100
   if (nscans ge 2 and (not keyword_set(nocross)) and keyword_set(simul_excess)) then begin
      otherplot, /over, time_save(w), drifts_save_init(w) + 4. * dev, psym=4, color=150
      otherplot, /over, time_save(w), drifts_save_init(w) - drifts_save(w) + 4. * dev, psym=4, color=200
   endif
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif



print, '              drifts amplitude / noise (median, min and max) = ' $
   + chain(median(amp_drifts / maxnoise), 2) + ' ; ' $
   + chain(min(amp_drifts / maxnoise, /nan), 2) + ' to ' + chain(max(amp_drifts / maxnoise, /nan), 2) + crs



;;; update signal and total drifts:

if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   res = dialog_message('Subtract average drift ?', /center, /question)
   time_1 = systime(/seconds)
endif else res = 'Yes'

if res eq 'Yes' then begin
;;;
if (not keyword_set(separate_matrices)) then begin
   signal -= rebin(drifts, nt, nb)
   drifts_all += rebin(drifts_save, nt_save, nb)
   save, file=file_signal, signal, /xdr
endif else begin
   restore, file_signal_orig
   signal_orig(*, wmat) -= rebin(drifts, nt, nb_mat)
   save, file=file_signal_orig, signal_orig, /xdr
   signal_orig = 0
   drifts_all(*, wmat) += rebin(drifts_save, nt_save, nb_mat)
endelse
;;;

   ;;; visualization of drift:
   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
;;;
if keyword_set(separate_matrices) then begin
   signal = signal_save
   signal(*, wmat) -= rebin(drifts, nt, nb_mat)
   save, file=file_signal, signal, /xdr
endif
;;;
      proj_map_artemis, map_nodrifts, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, $
         ind_subscans=ind_subscans
;;;
if keyword_set(separate_matrices) then begin
   signal = signal_save
   save, file=file_signal, signal, /xdr
endif
;;;
      title = 'map of iteration ' + chain(iter) + ' drifts'
      if keyword_set(detect_gl) then title = title + ' (+ glitches)'
      map_drifts = map_signal - map_nodrifts
      disp_ima, map_drifts, win=3, title=title, weightmap=map_weight, noise_level=median(maxnoise)
      map_ratio = fltarr(dimx, dimy)
      w = where(map_signal gt median(maxnoise))
      map_ratio(w) = abs(map_drifts(w)) / map_signal(w)
      disp_ima, map_ratio, win=4, title='ratio map of iteration ' + chain(iter) + ' |drifts| / signal', $
         weightmap=map_weight, noise_level=median(maxnoise), max_map=min([max(map_ratio), 0.15])
      if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
endif else amp_drifts *= 0.


;;;
if keyword_set(separate_matrices) then amp_drifts_all += amp_drifts^2.

endfor
if keyword_set(separate_matrices) then begin
   restore, file_signal_orig
   signal = signal_orig
   signal_orig = 0
   save, file=file_signal, signal, /xdr
   maxnoise = maxnoise_orig
   highfnoise = highfnoise_orig
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   prepare_proj_coarse, file_scalars=file_scalars
   map_signal = map_signal_orig
   map_weight = map_weight_orig
   save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, map_signal_uncl, /xdr
   amp_drifts = sqrt(amp_drifts_all / float(nmat))
endif
;;;

return, amp_drifts

end
