pro stat_map, med, sig, map=map, weightmap=weightmap, noise_level=noise_level


;;; estimates an approximate background level and dispersion of an image
;;; (used to define source masks)

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



w = where(weightmap gt 0. and finite(map) gt 0)
med_weight = median(weightmap(w))
sig_weight = max([sqrt(mean((weightmap(w) - med_weight)^2.)), 0.1 * med_weight])
w = where(weightmap gt 0. and finite(map) gt 0 and abs(weightmap - med_weight) lt 1.5 * sig_weight)
sig_weight = max([sqrt(mean((weightmap(w) - med_weight)^2.)), 0.1 * med_weight])
thres_weight = max([med_weight - 2.5 * sig_weight, 0.01 * med_weight])
thres_weight = min([0.4 * med_weight, thres_weight])
w = where(weightmap gt thres_weight and finite(map) gt 0, cw)
ind_sort = sort(map(w))
sb_min = map(w(ind_sort(cw/1000)))
sb_max = map(w(ind_sort(cw-1-cw/1000)))
w = where(weightmap gt thres_weight and finite(map) gt 0 and map ge sb_min and map le sb_max)
binsize = max([noise_level / 3., (max(map(w)) - min(map(w))) / 1000.])
binsize = min([binsize, (max(map(w)) - min(map(w))) / 100.])
histo = histogram(map(w), locations=sb, binsize=binsize)
m = max(histo, ind)
w = where(histo gt m / 20., cw)
med = sb(w(0))
if sb(ind) - med le (Machar()).eps then lim = sb(w(cw-1)) else lim = sb(ind)
w = where(weightmap gt thres_weight and finite(map) gt 0 and map le lim)
sig = sqrt(mean((map(w) - med)^2.))
if sig lt 100. * noise_level then for k = 0, 9 do begin
   if sig gt (Machar()).eps then begin
      w = where(weightmap gt thres_weight and map - med le 2.5 * sig)
      binsize = max([noise_level / 3., (med + 2. * sig - min(map(w))) / 1000.])
      binsize = min([binsize, (med + 2. * sig - min(map(w))) / 100.])
      histo = histogram(map(w), locations=sb, binsize=binsize)
      m = max(histo, ind)
      w = where(histo gt m / 20., cw)
      med = sb(ind)
      if med - sb(w(0)) le (Machar()).eps then lim = sb(w(cw-1)) else lim = med
      w = where(weightmap gt thres_weight and map le lim)
      sig = sqrt(mean((map(w) - med)^2.))
      w = where(weightmap gt thres_weight and abs(map - med) le 2. * sig)
      med = median(map(w))
   endif else sig = noise_level / 3.
endfor else sig = 10. * noise_level

end
